from pathlib import Path

import loguru
from pinjected import *
import pinjected

from pinjected_openai.clients import async_openai_client, openai_api_key

__version__ = "0.4.24"

default_design = design(
    cache_root_path=Path("~/.cache/pinjected_openai").expanduser(),
) + providers(
    logger=lambda: loguru.logger,
    async_openai_client=async_openai_client,
    #openai_api_key=openai_api_key
)
__meta_design__ = instances(
    #default_design_path="pinjected_openai.default_design"
    overrides=default_design
)



from pinjected import *
from pydantic import BaseModel

from pinjected_openai.vision_llm import to_content
from openai.types.chat import ChatCompletion
from openai import AsyncOpenAI
from typing import Optional,Literal
from PIL.Image import Image
@injected
async def a_openai_compatible_llm(
        logger,
        /,
        api:AsyncOpenAI,
        model:str,
        text:str,
        images:Optional[list[Image]]=None,
        response_format=None,
        max_completion_tokens:int = None,
        reasoning_effort=None,
        detail:Literal["auto","low","high"] = "auto"
)->ChatCompletion:
    images = images or []

    api_kwargs = dict(
        messages=[
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": text},
                    *[to_content(img, detail=detail) for img in images],
                ],
            }
        ],
        model=model,
        response_format=response_format,
    )

    if max_completion_tokens is not None:
        api_kwargs['max_completion_tokens'] = max_completion_tokens
    if reasoning_effort is not None:
        api_kwargs['reasoning_effort'] = reasoning_effort
    if isinstance(response_format,type) and issubclass(response_format,BaseModel):
        api_kwargs['response_format'] = response_format.model_json_schema()
        api_object = api.beta.chat.completions.parse
    else:
        api_object = api.chat.completions.create

    return await api_object(**api_kwargs)




import os
from pathlib import Path

from openai import AsyncOpenAI
from pinjected import instance


@instance
def async_openai_client(openai_api_key, openai_organization) -> AsyncOpenAI:
    return AsyncOpenAI(
        api_key=openai_api_key,
        organization=openai_organization
    )


@instance
def openai_api_key() -> str:
    from loguru import logger
    logger.warning(f"using openai api key from environment variable or ~/.openai_api_key.txt")
    if (api_key := os.environ.get('OPENAI_API_KEY', None)) is None:
        api_key = Path(os.path.expanduser("~/.openai_api_key.txt")).read_text().strip()
    return api_key



import asyncio
from io import BytesIO
from pathlib import Path

import pydub
from beartype import beartype
from injected_utils.injected_cache_utils import sqlite_cache, async_cached
from openai import AsyncOpenAI
from openai.types.audio import Transcription
from pinjected import *
from tqdm import tqdm


def convert_mp4_to_mp3(input_file: str, output_file: str = None) -> str:
    """
    Convert an MP4 file to MP3 format.

    Args:
    input_file (str): Path to the input MP4 file.
    output_file (str, optional): Path for the output MP3 file. If not provided,
                                 it will use the same name as the input file
                                 with an .mp3 extension.

    Returns:
    str: Path to the output MP3 file.

    Raises:
    FileNotFoundError: If the input file doesn't exist.
    RuntimeError: If there's an error during the conversion process.
    """
    import moviepy.editor as mp
    input_path = Path(input_file)

    if not input_path.exists():
        raise FileNotFoundError(f"Input file not found: {input_file}")

    if output_file is None:
        output_file = input_path.with_suffix('.mp3')
    else:
        output_file = Path(output_file)

    try:
        # Try to load as video first
        try:
            clip = mp.VideoFileClip(str(input_path))
        except Exception as video_error:
            # If loading as video fails, try to load as audio
            try:
                clip = mp.AudioFileClip(str(input_path))
            except Exception as audio_error:
                raise RuntimeError(f"Failed to load file as video ({str(video_error)}) or audio ({str(audio_error)})")

        # Extract audio (or use the audio directly if it's an audio file)
        audio = clip.audio if hasattr(clip, 'audio') else clip

        # Write audio to file
        audio.write_audiofile(str(output_file))

        # Close the clips
        audio.close()
        if audio != clip:
            clip.close()

        return str(output_file)
    except Exception as e:
        raise RuntimeError(f"Error converting {input_file} to MP3: {str(e)}")


@injected
@beartype
async def a_transcribe_bytes(async_openai_client: AsyncOpenAI,logger, /, sound_bytes: BytesIO) -> str:
    response: Transcription = await async_openai_client.audio.transcriptions.create(
        model="whisper-1",
        file=sound_bytes,
    )
    return response.text


@injected
async def a_transcribe_audio_segment(
        a_transcribe_bytes,
        a_split_segment_to_chunks,
        logger,
        /,
        segment: pydub.AudioSegment,
        bitrate="96k"
) -> str:
    transcription = ""
    bar = tqdm(desc="Transcribing chunks")
    i = 0
    from pinjected.compatibility.task_group import TaskGroup
    async def task(i, buffer):
        logger.info(f"Transcribing chunk of size {len(byte_segment.read()) / 1024 / 1024:.2f} MB")
        text = await a_transcribe_bytes(sound_bytes=buffer)
        logger.info(f"Transcribed chunk {i}: {text}")
        bar.update(1)
        return text

    async with TaskGroup() as tg:
        tasks = []
        async for byte_segment in a_split_segment_to_chunks(segment, bitrate=bitrate):
            tasks.append(tg.create_task(task(i, byte_segment)))
            i += 1
    for task in tasks:
        transcription += await task
    bar.close()
    return transcription


@injected
async def a_split_segment_to_chunks(
        logger, /,
        segment: pydub.AudioSegment,
        chunk_size_mb: float = 5,
        bitrate="96k",
):
    """
    recursively try to split the segment into chunks of size chunk_size_mb
    """
    from io import BytesIO
    def export(segment):
        out_bytes = BytesIO()
        out_bytes.name = "exported.mp3"
        segment.export(out_f=out_bytes, format="mp3", bitrate=bitrate)
        out_bytes.seek(0)
        size = len(out_bytes.read())
        out_bytes.seek(0)
        return out_bytes, size / 1024 / 1024

    async def impl(segment):
        assert isinstance(segment, pydub.AudioSegment)
        exported, size_mb = await asyncio.get_running_loop().run_in_executor(None,export,segment)
        if size_mb <= chunk_size_mb:
            sec = len(segment) / 1000
            logger.info(f"Exported chunk of size {size_mb:.2f} MB, duration {sec:.2f} sec")
            yield exported
        else:
            logger.info(f"Splitting chunk of size {size_mb:.2f} MB, duration {len(segment) / 1000:.2f} sec")
            half = len(segment) // 2
            async for item in impl(segment[:half]):
                yield item
            async for item in impl(segment[half:]):
                yield item

    async for item in impl(segment):
        yield item


# @async_cached(
#     cache=sqlite_cache(injected("cache_root_path") / "transcribe_mp3_file.sqlite"),
# )
@injected
async def a_transcribe_mp3_file(
        a_transcribe_audio_segment,
        /,
        file: Path,
        start_sec: float = None,
        end_sec: float = None) -> str:
    segment = await get_audio_segment(file)

    if start_sec is None:
        start_millis = 0
    else:
        start_millis = min(start_sec * 1000, len(segment) - 1)

    if end_sec is None:
        end_millis = len(segment) - 1
    else:
        end_millis = min(end_sec * 1000, len(segment) - 1)
    return await a_transcribe_audio_segment(segment[start_millis:end_millis])


async def get_audio_segment(file):
    file = Path(file)
    match file.suffix:
        case ".mp3":
            segment = pydub.AudioSegment.from_file(file, format="mp3")
        case ".mp4":
            mp3_file = file.with_suffix(".mp3")
            if not mp3_file.exists():
                convert_mp4_to_mp3(file, mp3_file)
            segment = pydub.AudioSegment.from_file(mp3_file, format="mp3")
        case ".wav":
            segment = pydub.AudioSegment.from_file(file, format="wav")
        case _:
            raise ValueError(f"Unsupported file type {file.suffix}")
    return segment


test_transcribe: IProxy = a_transcribe_mp3_file(Path("recording.mp3"))


@injected
async def __save_text(text: str, path: Path):
    path.write_text(text)


cmd_save_transcribe = __save_text(
    a_transcribe_mp3_file(injected("input_file")),
    injected("input_file").eval().map(Path).proxy.with_suffix(".txt")
)
with design(
        input_file='/Users/s22625/Downloads/2024.10.15 13.56 Backtrack.mp4'
):
    test_transcribe_mp4: IProxy = cmd_save_transcribe

__meta_design__ = instances(
)



import asyncio
import base64
import io
import json
import random
import re
from asyncio import Lock
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Literal, Callable, Optional, Awaitable

import openai.types.chat
import pandas as pd
import reactivex
from PIL.Image import Image
from injected_utils.injected_cache_utils import async_cached, sqlite_dict
from loguru import logger
from math import ceil
from openai import AsyncOpenAI, RateLimitError, APITimeoutError, APIConnectionError
from openai.types.chat import ChatCompletion
from pydantic import BaseModel
from pydantic import Field
from pinjected import *
from tenacity import retry,stop_after_attempt,retry_if_exception_type

class ChatCompletionWithCost(BaseModel):
    src: ChatCompletion
    total_cost_usd: float
    prompt_cost_usd: float
    completion_cost_usd: float


@instance
def chat_completion_costs_subject():
    return reactivex.Subject()


def to_content(img: Image, detail: Literal["auto", "low", "high"] = "auto"):
    # convert Image into jpeg bytes
    jpg_bytes = io.BytesIO()
    img.convert("RGB").save(jpg_bytes, format="jpeg", quality=95)
    b64_image = base64.b64encode(jpg_bytes.getvalue()).decode("utf-8")
    mb_of_b64 = len(b64_image) / 1024 / 1024
    logger.info(f"image size: {mb_of_b64:.2f} MB in base64.")
    return {
        "type": "image_url",
        "image_url": dict(url=f"data:image/jpeg;base64,{b64_image}", detail=detail),
    }


@dataclass
class UsageEntry:
    timestamp: datetime
    tokens: int

    class Config:
        arbitrary_types_allowed = True


class RateLimitManager(BaseModel):
    max_tokens: int
    max_calls: int
    duration: timedelta
    lock: Lock = asyncio.Lock()
    call_history: list[UsageEntry] = []

    async def acquire(self, approx_tokens):
        if await self.ready(approx_tokens):
            pass
        else:
            # wait for some time or condition, but who checks the condition?
            # a distinct loop, or loop here?
            # 1. check if we need to wait
            # 2. check if someone else is waiting with loop
            # 3. if not use looping to wait
            # Currently, everyone waits with loops
            while not await self.ready(approx_tokens):
                await asyncio.sleep(1)

    async def ready(self, token):
        async with self.lock:
            remaining = await self.remaining_tokens()
            is_ready = remaining >= token and len(self.call_history) < self.max_calls
            if is_ready:
                self.call_history.append(UsageEntry(pd.Timestamp.now(), token))
            return is_ready

    async def remaining_tokens(self):
        return self.max_tokens - await self._current_usage()

    async def remaining_calls(self):
        return self.max_calls - len(self.call_history)

    async def _current_usage(self):
        t = pd.Timestamp.now()
        self.call_history = [
            e for e in self.call_history if e.timestamp > t - self.duration
        ]
        return sum(e.tokens for e in self.call_history)

    class Config:
        arbitrary_types_allowed = True


class RateLimitKey(BaseModel):
    api_key: str
    organization: str
    model_name: str
    request_type: str


class BatchQueueLimits(BaseModel):
    tpm: int = Field(None, alias="TPM")
    rpm: int = Field(None, alias="RPM")
    tpd: int = Field(None, alias="TPD")
    images_per_minute: int = None


class ModelLimits(BaseModel):
    modeltoken_limits: int = None
    request_limits: int = None
    other_limits: int = None
    batch_queue_limits: BatchQueueLimits


class Limits(BaseModel):
    gpt_4: ModelLimits = Field(None, alias="gpt-4")
    gpt_3_5_turbo: ModelLimits = Field(None, alias="gpt-3.5-turbo")
    gpt_4_turbo: ModelLimits = Field(None, alias="gpt-4-turbo")
    text_embedding_3_small: ModelLimits = Field(None, alias="text-embedding-3-small")
    dall_e_3: ModelLimits = Field(None, alias="dall-e-3")
    tts_1: ModelLimits = Field(None, alias="tts-1")
    whisper_1: ModelLimits = Field(None, alias="whisper-1")


personal_limits = Limits(
    gpt_4=ModelLimits(
        modeltoken_limits=10000000,
        request_limits=10000,
        other_limits=1500000000,
        batch_queue_limits=BatchQueueLimits(tpm=300000, rpm=10000, tpd=45000000),
    ),
    gpt_3_5_turbo=ModelLimits(
        modeltoken_limits=2000000,
        request_limits=10000,
        other_limits=300000000,
        batch_queue_limits=BatchQueueLimits(tpm=2000000, rpm=10000, tpd=300000000),
    ),
    gpt_4_turbo=ModelLimits(
        modeltoken_limits=2000000,
        request_limits=10000,
        other_limits=300000000,
        batch_queue_limits=BatchQueueLimits(tpm=2000000, rpm=10000, tpd=300000000),
    ),
    text_embedding_3_small=ModelLimits(
        modeltoken_limits=10000000,
        request_limits=10000,
        other_limits=4000000000,
        batch_queue_limits=BatchQueueLimits(tpm=10000000, rpm=10000, tpd=4000000000),
    ),
    dall_e_3=ModelLimits(batch_queue_limits=BatchQueueLimits(images_per_minute=50)),
    tts_1=ModelLimits(batch_queue_limits=BatchQueueLimits(rpm=500)),
    whisper_1=ModelLimits(batch_queue_limits=BatchQueueLimits(rpm=500)),
)


class ModelPricing(BaseModel):
    input_cost: float
    output_cost: float


class PricingModel(BaseModel):
    gpt_4_turbo: ModelPricing = ModelPricing(input_cost=0.0100, output_cost=0.0300)
    gpt_4_turbo_2024_04_09: ModelPricing = ModelPricing(
        input_cost=0.0100, output_cost=0.0300
    )
    gpt_4: ModelPricing = ModelPricing(input_cost=0.0300, output_cost=0.0600)
    gpt_4_32k: ModelPricing = ModelPricing(input_cost=0.0600, output_cost=0.1200)
    gpt_4_0125_preview: ModelPricing = ModelPricing(
        input_cost=0.0100, output_cost=0.0300
    )
    gpt_4_1106_preview: ModelPricing = ModelPricing(
        input_cost=0.0100, output_cost=0.0300
    )
    gpt_4_vision_preview: ModelPricing = ModelPricing(
        input_cost=0.0100, output_cost=0.0300
    )
    gpt_3_5_turbo_1106: ModelPricing = ModelPricing(
        input_cost=0.0010, output_cost=0.0020
    )
    gpt_3_5_turbo_0613: ModelPricing = ModelPricing(
        input_cost=0.0015, output_cost=0.0020
    )
    gpt_3_5_turbo_16k_0613: ModelPricing = ModelPricing(
        input_cost=0.0030, output_cost=0.0040
    )
    gpt_3_5_turbo_0301: ModelPricing = ModelPricing(
        input_cost=0.0015, output_cost=0.0020
    )
    davinci_002: ModelPricing = ModelPricing(input_cost=0.0020, output_cost=0.0020)
    babbage_002: ModelPricing = ModelPricing(input_cost=0.0004, output_cost=0.0004)
    gpt_4o_mini_2024_07_18:ModelPricing = ModelPricing(input_cost=0.15*0.001, output_cost=0.6*0.001)
    gpt_4o: ModelPricing = ModelPricing(input_cost=0.0025, output_cost=0.0150)
    gpt_4o_2024_05_13: ModelPricing = ModelPricing(
        input_cost=0.0025, output_cost=0.0150
    )
    gpt_4o_2024_08_06: ModelPricing = ModelPricing(
        input_cost=0.0025, output_cost=0.0150
    )
    # o3_mini_2025_01_31: ModelPricing = ModelPricing(input_cost=1.10/1000, output_cost=4.40/1000)
    o3_mini_2025_01_31: ModelPricing = ModelPricing(
        input_cost=0.0011, output_cost=0.0044
    )
    o1_2024_12_17: ModelPricing = ModelPricing(input_cost=0.0151, output_cost=0.060)


pricing_model = PricingModel()


@instance
def openai_rate_limit_managers(
    openai_api_key, openai_organization, openai_rate_limits: Limits
) -> dict[Any, RateLimitManager]:
    managers = dict()
    for model, limits in openai_rate_limits.dict().items():
        key = RateLimitKey(
            api_key=openai_api_key,
            organization=openai_organization,
            model_name=model,
            request_type="completion",
        )
        managers[key] = RateLimitManager(
            max_tokens=limits.modeltoken_limits,
            max_counts=limits.request_limits,
            duration=pd.Timedelta("1 minute"),
        )
    return managers


@injected
async def a_repeat_for_rate_limit(logger, /, task):
    default_wait = 10
    while True:
        try:
            return await task()
        except RateLimitError as e:
            logger.error(f"rate limit error: {e}")
            pat = r"Please try again in (\d+) seconds."
            match = re.search(pat, e.message)
            if match:
                seconds = int(match.group(1))
                logger.info(f"sleeping for {seconds} seconds")
                await asyncio.sleep(seconds)
            else:
                logger.warning(f"failed to parse rate limit error message: {e.message}")
                zitter = random.random() * 30
                to_wait = default_wait + zitter
                await asyncio.sleep(to_wait)
                default_wait = to_wait
        except APITimeoutError as e:
            logger.warning(f"API timeout error: {e}")
            await asyncio.sleep(10)
        except APIConnectionError as ace:
            logger.warning(f"API connection error: {ace}")
            await asyncio.sleep(10)


def resize(width, height):
    if width > 1024 or height > 1024:
        if width > height:
            height = int(height * 1024 / width)
            width = 1024
        else:
            width = int(width * 1024 / height)
            height = 1024
    return width, height


@injected
def openai_count_image_tokens(width: int, height: int):
    width, height = resize(width, height)
    h = ceil(height / 512)
    w = ceil(width / 512)
    total = 85 + 170 * h * w
    return total


@injected
async def a_chat_completion_to_cost(
    openai_model_pricing_table: dict[str, ModelPricing], /, completion: ChatCompletion
) -> ChatCompletionWithCost:
    pricing = openai_model_pricing_table[completion.model]
    usage = completion.usage
    return ChatCompletionWithCost(
        src=completion,
        total_cost_usd=pricing.input_cost * usage.prompt_tokens / 1000
        + pricing.output_cost * usage.completion_tokens / 1000,
        prompt_cost_usd=pricing.input_cost * usage.prompt_tokens / 1000,
        completion_cost_usd=pricing.output_cost * usage.completion_tokens / 1000,
    )


@instance
def openai_model_pricing_table(logger):
    keys = pricing_model.dict().keys()
    # keys = [k.replace("_", "-") for k in keys]
    table = {k.replace("_", "-"): getattr(pricing_model, k) for k in keys}
    logger.info(f"openai api pricing table:{table}")
    return table


"""
TODO rate limit using model_name,
and retry exponentially.
"""


class StructuredLLMNoneException(Exception):
    def __init__(self, message, completion):
        super().__init__(message)
        self.completion = completion


@instance
def __openai_call_stat__():
    return dict()

class StructuredLLMRefusalException(Exception):
    def __init__(self, message, completion):
        super().__init__(message)
        self.completion = completion

@injected
async def a_call_openai_api(
    a_repeat_for_rate_limit,
    a_enable_cost_logging,
    a_chat_completion_to_cost,
    __openai_call_stat__,
    /,
    api_object: Callable[[...], Awaitable],
    api_kwargs: dict,
) -> ChatCompletionWithCost:
    await a_enable_cost_logging()
    async def task():
        __openai_call_stat__["total_calls"] = (
            __openai_call_stat__.get("total_calls", 0) + 1
        )
        chat_completion = await api_object(**api_kwargs)
        cost: ChatCompletionWithCost = await a_chat_completion_to_cost(chat_completion)
        chat_completion_costs_subject.on_next(cost)
        __openai_call_stat__["total_cost_usd"] = (
            __openai_call_stat__.get("total_cost_usd", 0) + cost.total_cost_usd
        )
        cumulative_cost = __openai_call_stat__["total_cost_usd"]
        logger.info(
            f"cost: {cost.total_cost_usd:.4f} USD, cumulative: {cumulative_cost:.4f} USD"
        )
        refusal = cost.src.choices[0].message.refusal
        if refusal is not None:
            raise StructuredLLMRefusalException(f"refusal from llm:{refusal}", cost)
        return cost

    return await a_repeat_for_rate_limit(task)


@injected
async def a_vision_llm__openai(
    async_openai_client: AsyncOpenAI,
    logger,
    a_call_openai_api,
    /,
    text: str,
    images: Optional[list[Image]] = None,
    model: str = "gpt-4o",
    max_tokens=None,  # deprecated
    max_completion_tokens=None,  # same as max_tokens
    response_format: openai.types.chat.completion_create_params.ResponseFormat = None,
    detail: Literal["auto", "low", "high"] = "auto",
    reasoning_effort: Literal["low", "medium", "high"] = None,
) -> str:
    if max_completion_tokens is None:
        if max_tokens is None:
            max_completion_tokens = 2048
        max_completion_tokens = max_tokens

    assert isinstance(async_openai_client, AsyncOpenAI)
    if images is None:
        images = []
    if response_format is None:
        response_format = {"type": "text"}

    for img in images:
        assert isinstance(img, Image), f"image is not Image, but {type(img)}"

    if isinstance(response_format, type) and issubclass(response_format, BaseModel):
        API = async_openai_client.beta.chat.completions.parse

        def get_result(completion):
            data = completion.choices[0].message.parsed
            if data is None:
                raise StructuredLLMNoneException("data from llm is None", completion)
            logger.info(completion)
            return data

    else:
        API = async_openai_client.chat.completions.create

        def get_result(completion):
            return completion.choices[0].message.content

    api_kwargs = dict(
        messages=[
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": text},
                    *[to_content(img, detail=detail) for img in images],
                ],
            }
        ],
        model=model,
        max_completion_tokens=max_completion_tokens,
        response_format=response_format,
    )
    if reasoning_effort is not None:
        api_kwargs["reasoning_effort"] = reasoning_effort

    @retry(
        retry=retry_if_exception_type(StructuredLLMRefusalException),
        stop=stop_after_attempt(3),
    )
    async def task():
        chat_completion: ChatCompletionWithCost = await a_call_openai_api(
            api_object=API, api_kwargs=api_kwargs
        )
        return get_result(chat_completion.src)
    return await task()


@instance
async def cost_logging_state():
    return dict(
        enabled=False,
    )


@injected
async def a_enable_cost_logging(
    cost_logging_state: dict, chat_completion_costs_subject: reactivex.Subject, /
):
    if cost_logging_state["enabled"]:
        return
    cumulative_cost = 0

    def on_next(cost: ChatCompletionWithCost):
        nonlocal cumulative_cost
        cumulative_cost += cost.total_cost_usd
        logger.info(
            f"cost: {cost.total_cost_usd:.4f} USD, cumulative: {cumulative_cost:.4f} USD"
        )

    chat_completion_costs_subject.subscribe(on_next)
    cost_logging_state["enabled"] = True


a_vision_llm__gpt4o = Injected.partial(a_vision_llm__openai, model="gpt-4o")
_test_a_gpt4o: Injected = Injected.procedure(
    a_enable_cost_logging(),
    a_vision_llm__gpt4o("hello?"),
    a_vision_llm__gpt4o("hello hello"),
)
a_vision_llm__gpt4 = Injected.partial(
    a_vision_llm__openai, model="gpt-4-vision-preview"
)
a_cached_vision_llm__gpt4o = async_cached(
    sqlite_dict(
        str(Path("~/.cache/pinjected_openai/a_vision_llm__gpt4o.sqlite").expanduser())
    )
)(a_vision_llm__gpt4o)
a_cached_vision_llm__gpt4 = async_cached(
    sqlite_dict(
        str(Path("~/.cache/pinjected_openai/a_vision_llm__gpt4.sqlite").expanduser())
    )
)(a_vision_llm__gpt4)


@injected
async def a_llm__openai(
    async_openai_client: AsyncOpenAI,
    a_repeat_for_rate_limit,
    a_enable_cost_logging,
    /,
    text: str,
    model_name: str,
    max_completion_tokens=4096,
) -> str:
    assert isinstance(async_openai_client, AsyncOpenAI)
    await a_enable_cost_logging()

    async def task():
        # import tiktoken
        # enc = tiktoken.get_encoding("cl100k_base")
        # n_token = len(enc.encode(text))
        chat_completion = await async_openai_client.chat.completions.create(
            messages=[
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": text},
                    ],
                }
            ],
            model=model_name,
            max_tokens=max_completion_tokens,
        )
        return chat_completion

    chat_completion = await a_repeat_for_rate_limit(task)
    res = chat_completion.choices[0].message.content
    assert isinstance(res, str)
    logger.info(f"result:\n{res}")
    return res


@injected
async def a_llm__gpt4_turbo(
    a_llm__openai, /, text: str, max_completion_tokens=4096
) -> str:
    return await a_llm__openai(
        text,
        max_completion_tokens=max_completion_tokens,
        model_name="gpt-4-turbo-preview",
    )


a_llm__gpt4_turbo_cached = async_cached(
    sqlite_dict(str(Path("~/.cache/a_llm__gpt4_turbo.sqlite").expanduser()))
)(a_llm__gpt4_turbo)


@injected
async def a_llm__gpt35_turbo(
    a_llm__openai, /, text: str, max_completion_tokens=4096
) -> str:
    return await a_llm__openai(
        text, max_completion_tokens=max_completion_tokens, model_name="gpt-3.5-turbo"
    )


@injected
async def a_json_llm__openai(
    logger,
    async_openai_client: AsyncOpenAI,
    a_repeat_for_rate_limit,
    /,
    text: str,
    max_completion_tokens=4096,
    model="gpt-4-0125-preview",
) -> str:
    assert isinstance(async_openai_client, AsyncOpenAI)

    async def task():
        # import tiktoken
        # enc = tiktoken.get_encoding("cl100k_base")
        # n_token = len(enc.encode(text))
        chat_completion = await async_openai_client.chat.completions.create(
            messages=[
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": text},
                    ],
                }
            ],
            model=model,
            max_tokens=max_completion_tokens,
            response_format={"type": "json_object"},
        )
        return chat_completion

    chat_completion = await a_repeat_for_rate_limit(task)
    res = chat_completion.choices[0].message.content
    assert isinstance(res, str)
    logger.info(f"\n{res}")
    return json.loads(res)


@injected
async def a_json_llm__gpt4_turbo(
    a_json_llm__openai, /, text: str, max_completion_tokens=4096
) -> str:
    return await a_json_llm__openai(
        text=text,
        max_completion_tokens=max_completion_tokens,
        model="gpt-4-turbo-preview",
    )


@injected
async def a_vlm__openai_batched(
    async_openai_client: AsyncOpenAI,
    a_repeat_for_rate_limit,
    a_enable_cost_logging,
    /,
    texts: list[str],
    model_name: str,
    max_completion_tokens=4096,
) -> list[str]:
    assert isinstance(async_openai_client, AsyncOpenAI)
    await a_enable_cost_logging()

    async def task():
        completions = await async_openai_client.chat.completions.create_batch(
            messages=[
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": text},
                    ],
                }
                for text in texts
            ],
            model=model_name,
            max_tokens=max_completion_tokens,
        )
        return [c.choices[0].message.content for c in completions]

    completions = await a_repeat_for_rate_limit(task)
    return completions


test_vision_llm__gpt4 = a_vision_llm__gpt4(
    text="What are inside this image?",
    images=Injected.list(),
)
"""
('The image appears to be an advertisement or an informational graphic about '
 'infant and newborn nutrition. It features a baby with light-colored hair who '
 'is lying down and holding onto a baby bottle, seemingly feeding themselves. '
 'The baby is looking directly towards the camera. The image uses a soft pink '
 'color palette, which is common for baby-related products or information. '
 'There are texts that read "Infant & Newborn Nutrition" and "Absolutely New," '
 'along with the word "PINGUIN" at the top, which could be a brand name or '
 "logo. The layout and design of this image suggest it's likely used for "
 'marketing purposes or as part of educational material regarding baby '
 'nutrition.')
"""

test_llm__gpt4_turbo = a_llm__gpt4_turbo("Hello world")

test_json_llm__gpt4_turbo = a_json_llm__gpt4_turbo("Hello world, respond to me in json")

__meta_design__ = instances()




import time
from typing import Protocol

from httpx import ReadTimeout
from injected_utils import async_cached, lzma_sqlite
from pinjected import injected, Injected, IProxy
from pydantic import BaseModel


class NoEndpointsFoundError(Exception):
    pass

class StructuredLLM(Protocol):
    async def __call__(self, text: str, model: str, images=None, response_format: type[BaseModel] = None, max_tokens: int = 8192):
        ...

@injected
async def a_sllm_openrouter(
        a_openrouter_chat_completion,
        logger,
        /,
        text: str,
        model: str,
        images=None,
        response_format: type[BaseModel] = None,
        max_tokens: int = 8192,
):
    retry_count = 5
    while retry_count:
        try:
            return await a_openrouter_chat_completion(
                text,
                response_format=response_format,
                # model='google/gemini-2.0-flash-001',
                images=images,
                model=model,
                max_tokens=max_tokens
            )
        except ReadTimeout as e:
            logger.warning(f"Timeout error for {a_openrouter_chat_completion.__name__}: {e}")
            time.sleep(1)
        except Exception as e:
            import traceback
            trc = traceback.format_exc()
            logger.warning(f"Remaining ({retry_count}) for {a_openrouter_chat_completion.__name__}: {e} \n {trc}")
            if 'No endpoints found' in str(e):
                raise NoEndpointsFoundError(f"No endpoints found for {model} with {response_format}") from e
            retry_count -= 1
            if not retry_count:
                raise e
            # if not retry_count:
            # raise RuntimeError(f"Failed to run {a_openrouter_chat_completion.__name__} after 5 retries.") from e


a_cached_structured_llm__gemini_flash_2_0: IProxy[StructuredLLM] = async_cached(
    lzma_sqlite(injected('cache_root_path') / "gemini_flash.sqlite"))(
    Injected.partial(a_sllm_openrouter, model='google/gemini-2.0-flash-001')
)
a_cached_structured_llm__deepseek_chat: IProxy[StructuredLLM] = async_cached(
    lzma_sqlite(injected('cache_root_path') / "deepseek_chat.sqlite"))(
    Injected.partial(a_sllm_openrouter, model='deepseek/deepseek-chat')
)
a_cached_structured_llm__gemini_flash_thinking_2_0: IProxy[StructuredLLM] = async_cached(
    lzma_sqlite(injected('cache_root_path') / "gemini_flash_thinking.sqlite"))(
    Injected.partial(a_sllm_openrouter, model='google/gemini-2.0-flash-thinking-exp:free')
)
a_cached_structured_llm__claude_sonnet_3_5: IProxy[StructuredLLM] = async_cached(
    lzma_sqlite(injected('cache_root_path') / "claude_sonnet_3_5.sqlite"))(
    Injected.partial(a_sllm_openrouter, model='anthropic/claude-3.5-sonnet')
)



import json_repair
from injected_utils.injected_cache_utils import sqlite_dict, async_cached
from openai.types import CompletionUsage
from pydantic import BaseModel
from openai import AsyncOpenAI
import httpx
from typing import Optional, List, Dict, Any, cast, Callable, Awaitable, Protocol
from pinjected_openai.compatibles import a_openai_compatible_llm
from pinjected import instance, design, IProxy, injected, Injected
import logging
from openai.types.chat import ChatCompletion, ParsedChoice
import PIL

from pinjected_openai.vision_llm import a_vision_llm__gpt4o, to_content


# from vision_llm import a_vision_llm__gpt4o


class OpenRouterArchitecture(BaseModel):
    modality: str
    tokenizer: str
    instruct_type: Optional[str] = None


class OpenRouterModelPricing(BaseModel):
    prompt: str
    completion: str
    image: str
    request: str

    def calc_cost(self, usage: CompletionUsage):
        completion_cost = usage.completion_tokens * float(self.completion)
        prompt_cost = usage.prompt_tokens * float(self.prompt)
        return dict(
            completion=completion_cost,
            prompt=prompt_cost,
        )

    def calc_cost_dict(self, usage: dict):
        completion_cost = usage['completion_tokens'] * float(self.completion)
        prompt_cost = usage['prompt_tokens'] * float(self.prompt)
        return dict(
            completion=completion_cost,
            prompt=prompt_cost,
        )


class OpenRouterTopProvider(BaseModel):
    context_length: Optional[int] = None
    max_completion_tokens: Optional[int] = None
    is_moderated: bool


class OpenRouterModel(BaseModel):
    id: str
    name: str
    created: int
    description: str
    context_length: int
    architecture: OpenRouterArchitecture
    pricing: OpenRouterModelPricing
    top_provider: Optional[OpenRouterTopProvider] = None
    per_request_limits: Optional[Dict[str, Any]] = None


class OpenRouterModelTable(BaseModel):
    data: List[OpenRouterModel]

    def pricing(self, model_id: str) -> OpenRouterModelPricing:
        if not hasattr(self, "_pricing"):
            self._pricing = {model.id: model.pricing for model in self.data}
        return self._pricing[model_id]


@instance
async def openrouter_model_table(logger) -> OpenRouterModelTable:
    async with httpx.AsyncClient() as client:
        response = await client.get("https://openrouter.ai/api/v1/models")
        response.raise_for_status()
        data = response.json()["data"]
        return OpenRouterModelTable.model_validate(dict(data=data))


@instance
def openrouter_api(openrouter_api_key: str):
    return AsyncOpenAI(
        base_url="https://openrouter.ai/api/v1", api_key=openrouter_api_key
    )


@instance
def openrouter_state():
    return dict()


@instance
def openrouter_timeout_sec() -> float:
    return 120


@injected
async def a_openrouter_post(
        openrouter_api_key: str,
        openrouter_timeout_sec: float,
        /,
        payload: dict
) -> dict:
    async with httpx.AsyncClient() as client:
        headers = {
            "Authorization": f"Bearer {openrouter_api_key}",
            'Content-Type': 'application/json',
        }
        response = await client.post("https://openrouter.ai/api/v1/chat/completions", headers=headers, json=payload,
                                     timeout=openrouter_timeout_sec)
        return response.json()


@async_cached(sqlite_dict(injected('cache_root_path') / "schema_examples.sqlite"))
@injected
async def a_cached_schema_example_provider(
        a_llm_for_json_schema_example,
        /,
        model_schema: dict
):
    prompt = f"""
    Provide example json objects that follows the schema of the model:{model_schema}
    Beware the example must not be in yaml format.
    If the model contains a list property, provide an example of a case where the list is empty and another example where the list contains multiple items.
    Beware that `type` field is required in the schema, so make sure to include it in the example.
    """
    return await a_llm_for_json_schema_example(prompt)


class OpenRouterChatCompletion(Protocol):
    async def __call__(
            self,
            prompt: str,
            model: str,
            max_tokens: int = 8192,
            temperature: float = 1,
            images: List[PIL.Image.Image] = None,
            response_format: Optional[BaseModel] = None,
            provider: Optional[Dict[str, Any]] = None,
            **kwargs
    ) -> Any:
        ...


@injected
async def a_openrouter_chat_completion__without_fix(
        a_openrouter_post,
        logger,
        openrouter_model_table: OpenRouterModelTable,
        openrouter_state: dict,
        /,
        prompt: str,
        model: str,
        max_tokens: int = 8192,
        temperature: float = 1,
        images: list[PIL.Image.Image] = None,
        response_format=None,
        provider: dict = None,
        **kwargs
):
    """
    :param prompt:
    :param model:
    :param max_tokens:
    :param temperature:
    :param images:
    :param response_format:
    :param provider:  see:https://openrouter.ai/docs/features/provider-routing
    Example:
    provider={'order': [
        'openai',
        'together'
      ],
      allow_fallbacks=False # if everything in order fails, fails the completion. Default is True so some other provider will be used.
    }
    :param kwargs:
    :return:
    """
    provider_filter = dict()
    if response_format is not None and issubclass(response_format, BaseModel):
        provider_filter['provider'] = {
            "require_parameters": True
        }
        schema = response_format.model_json_schema()
        schema['strict'] = True
        provider_filter['response_format'] = dict(
            type='json_object',
            json_schema=schema
        )
    if provider is not None:
        p = provider_filter.get('provider', dict())
        p.update(provider)
        provider_filter['provider'] = p

    payload: Dict[str, Any] = {
        "model": model,
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    # *[{"type": "image", "data": img} for img in images or []]
                    *[to_content(img) for img in images or []]
                ]
            }
        ],
        "max_tokens": max_tokens,
        "temperature": temperature,
        **provider_filter,
        **kwargs
    }
    from pprint import pformat
    res = await a_openrouter_post(payload)
    if 'error' in res:
        raise RuntimeError(f"Error in response: {pformat(res)}")
    cost_dict = openrouter_model_table.pricing(model).calc_cost_dict(res['usage'])
    openrouter_state['cumulative_cost'] = openrouter_state.get('cumulative_cost', 0) + sum(cost_dict.values())

    logger.info(
        f"Cost of completion: {cost_dict}, cumulative cost: {openrouter_state['cumulative_cost']} from {res['provider']}")
    data = res['choices'][0]['message']['content']

    if response_format is not None and issubclass(response_format, BaseModel):
        try:
            if '```' in data:
                data = data.split('```')[1].strip()
            return response_format.model_validate_json(data)
        except Exception as e:
            logger.warning(f"Error in response validation:\n{pformat(payload)}\n{pformat(res)} \n {e} cause:\n{data}")
            data_dict = json_repair.loads(data)
            return response_format.model_validate(data_dict)
    else:
        return data


@injected
async def a_openrouter_chat_completion(
        a_openrouter_post,
        logger,
        a_cached_schema_example_provider: Callable[[type], Awaitable[str]],
        a_structured_llm_for_json_fix,
        openrouter_model_table: OpenRouterModelTable,
        openrouter_state: dict,
        /,
        prompt: str,
        model: str,
        max_tokens: int = 8192,
        temperature: float = 1,
        images: list[PIL.Image.Image] = None,
        response_format=None,
        provider: dict = None,
        **kwargs
):
    """
    :param prompt:
    :param model:
    :param max_tokens:
    :param temperature:
    :param images:
    :param response_format:
    :param provider:  see:https://openrouter.ai/docs/features/provider-routing
    Example:
    provider={'order': [
        'openai',
        'together'
      ],
      allow_fallbacks=False # if everything in order fails, fails the completion. Default is True so some other provider will be used.
    }
    :param kwargs:
    :return:
    """
    provider_filter = dict()
    if response_format is not None and issubclass(response_format, BaseModel):
        provider_filter['provider'] = {
            "require_parameters": True
        }
        schema = response_format.model_json_schema()
        schema['strict'] = True
        provider_filter['response_format'] = dict(
            type='json_object',
            json_schema=schema
        )
        schema_prompt = await a_cached_schema_example_provider(response_format.model_json_schema())
        prompt += f"""The response must follow the following json format example:{schema_prompt}"""
    if provider is not None:
        p = provider_filter.get('provider', dict())
        p.update(provider)
        provider_filter['provider'] = p

    payload: Dict[str, Any] = {
        "model": model,
        "messages": [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    # *[{"type": "image", "data": img} for img in images or []]
                    *[to_content(img) for img in images or []]
                ]
            }
        ],
        "max_tokens": max_tokens,
        "temperature": temperature,
        **provider_filter,
        **kwargs
    }
    from pprint import pformat
    # logger.debug(f"payload:{pformat(payload)}")
    res = await a_openrouter_post(payload)
    if 'error' in res:
        raise RuntimeError(f"Error in response: {pformat(res)}")
    cost_dict = openrouter_model_table.pricing(model).calc_cost_dict(res['usage'])
    openrouter_state['cumulative_cost'] = openrouter_state.get('cumulative_cost', 0) + sum(cost_dict.values())
    # logger.debug(f"response:{pformat(res)}")

    logger.info(
        f"Cost of completion: {cost_dict}, cumulative cost: {openrouter_state['cumulative_cost']} from {res['provider']}")
    data = res['choices'][0]['message']['content']

    if response_format is not None and issubclass(response_format, BaseModel):
        try:
            if '```' in data:
                data = data.split('```')[1].strip()
            return response_format.model_validate_json(data)
        except Exception as e:
            logger.warning(f"Error in response validation:\n{pformat(payload)}\n{pformat(res)} \n {e} cause:\n{data}")
            try:
                data_dict = json_repair.loads(data)
                return response_format.model_validate(data_dict)
            except Exception as e:
                logger.warning(f"json_repair could not repair.{data}")
                fix_prompt = f"""
Please fix the following json object to match the schema:
{data}
                """
                return await a_structured_llm_for_json_fix(fix_prompt, response_format=response_format)
    else:
        return data


@injected
async def a_llm__openrouter(
        openrouter_model_table: OpenRouterModelTable,
        openrouter_api,
        a_openai_compatible_llm,
        logger,
        openrouter_state: dict,
        /,
        text: str,
        model: str,
        response_format=None,
        **kwargs
):
    res: ChatCompletion = await a_openai_compatible_llm(
        api=openrouter_api,
        model=model,
        text=text,
        response_format=response_format,
        **kwargs,
    )

    cost = openrouter_model_table.pricing(model).calc_cost(res.usage)
    total_cost = sum(cost.values())
    openrouter_state['cumulative_cost'] = openrouter_state.get('cumulative_cost', 0) + total_cost
    logger.info(
        f"Cost of completion: {cost}, total cost: {total_cost}, cumulative cost: {openrouter_state['cumulative_cost']}")

    data = res.choices[0].message.content
    if response_format is not None and issubclass(response_format, BaseModel):
        if '```' in data:
            data = data.split('```')[1].strip()
        data = response_format.model_validate_json(data)
    return data


class Text(BaseModel):
    text_lines: list[str]


test_call_gpt4o: IProxy = a_openrouter_chat_completion__without_fix(
    prompt="What is the capital of Japan?",
    model="openai/gpt-4o"
)

test_openai_compatible_llm: IProxy = a_openai_compatible_llm(
    api=openrouter_api,
    model="deepseek/deepseek-chat",
    text="What is the capital of Japan?",
)

test_openrouter_text: IProxy = a_llm__openrouter(
    "What is the capital of Japan?",
    "deepseek/deepseek-chat"
)

test_openrouter_structure: IProxy = a_llm__openrouter(
    f"What is the capital of Japan?.{Text.model_json_schema()}",
    # "deepseek/deepseek-chat",
    "deepseek/deepseek-r1-distill-qwen-32b",
    response_format=Text
)

test_openrouter_model_table: IProxy = openrouter_model_table

test_openrouter_chat_completion: IProxy = a_openrouter_chat_completion(
    prompt="What is the capital of Japan?",
    model="deepseek/deepseek-chat"
)

test_openrouter_chat_completion_with_structure: IProxy = a_openrouter_chat_completion(
    prompt=f"What is the capital of Japan?",
    model="deepseek/deepseek-chat",
    # model="deepseek/deepseek-r1-distill-qwen-32b",
    response_format=Text
)

test_return_empty_item: IProxy = a_openrouter_chat_completion(
    prompt=f"Please answer with empty lines.",
    model="deepseek/deepseek-chat",
    response_format=Text
)

a_cached_sllm_gpt4o__openrouter:IProxy = async_cached(sqlite_dict(injected('cache_root_path') / "gpt4o.sqlite"))(
    Injected.partial(
        a_openrouter_chat_completion__without_fix,
        model="openai/gpt-4o"
    )
)

a_cached_sllm_gpt4o_mini__openrouter: IProxy = async_cached(sqlite_dict(injected('cache_root_path') / "gpt4o_mini.sqlite"))(
    Injected.partial(
        a_openrouter_chat_completion__without_fix,
                     model="openai/gpt-4o-mini"
    )
)

test_cached_sllm_gpt4o_mini: IProxy = a_cached_sllm_gpt4o_mini__openrouter(
    prompt="What is the capital of Japan?",
    model="openai/gpt-4o-mini"
)
test_cached_sllm_gpt4o: IProxy = a_cached_sllm_gpt4o__openrouter(
    prompt="What is the capital of Japan?",
    model="openai/gpt-4o"
)

__meta_design__ = design(
    overrides=design(
        a_llm_for_json_schema_example=a_cached_sllm_gpt4o__openrouter,
        a_structured_llm_for_json_fix=a_cached_sllm_gpt4o_mini__openrouter,
        openai_config=injected('openai_config__personal')
    )
)
